import time

from confluent_kafka import Message
from confluent_kafka.serialization import StringDeserializer
from jlab_jaws.avro.serde import AlarmInstanceSerde
from jlab_jaws.eventsource.table import EventSourceTable
from jlab_jaws.eventsource.listener import EventSourceListener
from typing import List


class CacheListener(EventSourceListener):

    def __init__(self, parent: EventSourceTable):
        self._parent = parent

    def on_highwater(self):
        self._parent._highwater_signal.set()

    def on_highwater_timeout(self):
        pass

    def on_batch(self, msgs):
        self._parent.__update_cache(msgs)


class CachedTable(EventSourceTable):

    def __init__(self, config):
        self._cache = None

        super().__init__(config)

        self._listener = CacheListener(self)

        super().add_listener(self._listener)

    def __update_cache(self, msgs: List[Message]) -> None:
        for msg in msgs:
            if msg.value() is None:
                if msg.key() in self._state:
                    del self._state[msg.key()]
            else:
                self._state[msg.key()] = msg

    def await_get(self, timeout_seconds) -> List[Message]:
        """
        Synchronously get messages up to highwater mark.  Blocks with a timeout.

        :param timeout_seconds: Seconds to wait for highwater to be reached
        :return: List of Message
        :raises TimeoutException: If highwater is not reached before timeout
        """
        super().await_highwater(timeout_seconds)
        return self._cache


class CategoryCachedTable(CachedTable):
    def __init__(self, bootstrap_servers):
        key_deserializer = StringDeserializer('utf_8')
        value_deserializer = StringDeserializer('utf_8')

        ts = time.time()

        config = {'topic': 'alarm-categories',
                  'bootstrap.servers': bootstrap_servers,
                  'key.deserializer': key_deserializer,
                  'value.deserializer': value_deserializer,
                  'group.id': 'category-cached-table' + str(ts)}

        super().__init__(config)

        self.start()


class InstanceCachedTable(CachedTable):
    def __init__(self, bootstrap_servers, schema_registry_client):
        key_deserializer = StringDeserializer('utf_8')
        value_deserializer = AlarmInstanceSerde.deserializer(schema_registry_client)

        ts = time.time()

        config = {'topic': 'alarm-instances',
                  'bootstrap.servers': bootstrap_servers,
                  'key.deserializer': key_deserializer,
                  'value.deserializer': value_deserializer,
                  'group.id': 'instance-cached-table' + str(ts)}

        super().__init__(config)

        self.start()



"""
   Code to facilitate Event Sourcing.

   See Also:
       - `Storing Data in Kafka <https://www.confluent.io/blog/okay-store-data-apache-kafka/>`_
       - `Fowler on Event Sourcing <ttps://martinfowler.com/eaaDev/EventSourcing.html>`_
"""


"""A module for Event Sourcing
"""
from concurrent.futures import ThreadPoolExecutor
from typing import List

from confluent_kafka import DeserializingConsumer, OFFSET_BEGINNING, Message
from threading import Timer, Event
from jlab_jaws.eventsource.listener import EventSourceListener


class TimeoutException(Exception):
    pass


class EventSourceTable:
    """This class provides an Event Source Table abstraction.
    """

    __slots__ = ['_hash', '_config', '_listeners', '_state', '_consumer', '_executor',
                 '_end_reached', '_high', '_low', '_run', 'is_highwater_timeout']

    def __init__(self, config):
        """Create an EventSourceTable instance.

         Args:
             config (dict): Configuration
             on_initial_state (callable(dict): Callback providing initial state of EventSourceTable
             on_state_update (callable(dict)): Callback providing updated state

         Note:
             The configuration options include:

            +-------------------------+---------------------+-----------------------------------------------------+
            | Property Name           | Type                | Description                                         |
            +=========================+=====================+=====================================================+
            | ``bootstrap.servers``   | str                 | Comma-separated list of brokers.                    |
            +-------------------------+---------------------+-----------------------------------------------------+
            |                         |                     | Client group id string.                             |
            | ``group.id``            | str                 | All clients sharing the same group.id belong to the |
            |                         |                     | same group.                                         |
            +-------------------------+---------------------+-----------------------------------------------------+
            |                         |                     | Callable(SerializationContext, bytes) -> obj        |
            | ``key.deserializer``    | callable            |                                                     |
            |                         |                     | Deserializer used for message keys.                 |
            +-------------------------+---------------------+-----------------------------------------------------+
            |                         |                     | Callable(SerializationContext, bytes) -> obj        |
            | ``value.deserializer``  | callable            |                                                     |
            |                         |                     | Deserializer used for message values.               |
            +-------------------------+---------------------+-----------------------------------------------------+
            |                         |                     | Kafka topic name to consume messages from           |
            | ``topic``               | str                 |                                                     |
            |                         |                     |                                                     |
            +-------------------------+---------------------+-----------------------------------------------------+

            Note:
                Keys must be hashable so your key deserializer generally must generate immutable types.

         """
        self._config = config

        self._run = True
        self._low = None
        self._high = None
        self._state = {}

        self._is_highwater_timeout = False
        self._end_reached = False
        self._consumer = None
        self._executor = None
        self._highwater_signal = Event()
        self._listeners: List[EventSourceListener] = []

    def add_listener(self, listener: EventSourceListener):
        self._listeners.append(listener)

    def remove_listener(self, listener: EventSourceListener):
        self._listeners.remove(listener)

    def await_highwater(self, timeout_seconds: float) -> None:
        flag = self._highwater_signal.wait(timeout_seconds)
        if not flag:
            raise TimeoutException

    def start(self):
        """
            Start monitoring for state updates.
        """

        self._executor = ThreadPoolExecutor(max_workers=1)

        self._executor.submit(self.__monitor)

    def __do_highwater_timeout(self):
        self._is_highwater_timeout = True

    def __update_state(self, msg: Message):
        if msg.value() is None:
            if msg.key() in self._state:
                del self._state[msg.key()]
        else:
            self._state[msg.key()] = msg

    def __notify_changes(self):
        for listener in self._listeners:
            listener.on_batch(self._state.copy())

        self._state.clear()

    def __monitor(self):
        try:
            self.__monitor_initial()
            self.__monitor_continue()
        except Exception as e:
            print(type(e))
            print(e)
        finally:
            self._consumer.close()
            self._executor.shutdown()

    def __monitor_initial(self):
        consumer_conf = {'bootstrap.servers': self._config['bootstrap.servers'],
                         'key.deserializer': self._config['key.deserializer'],
                         'value.deserializer': self._config['value.deserializer'],
                         'group.id': self._config['group.id']}

        self._consumer = DeserializingConsumer(consumer_conf)
        self._consumer.subscribe([self._config['topic']], on_assign=self._my_on_assign)

        t = Timer(30, self.__do_highwater_timeout)
        t.start()

        while not (self._end_reached or self._is_highwater_timeout):
            msg = self._consumer.poll(1)

            msgs = [msg] if msg is not None else None

            if msgs is not None:
                for msg in msgs:
                    self.__update_state(msg)

                    if msg.offset() + 1 == self._high:
                        end_reached = True

                self.__notify_changes()

        t.cancel()

        if self._is_highwater_timeout:
            for listener in self._listeners:
                listener.on_highwater_timeout()
        else:
            for listener in self._listeners:
                listener.on_highwater()

    def __monitor_continue(self):
        while self._run:
            msg = self._consumer.poll(1)

            msgs = [msg] if msg is not None else None

            if msgs is not None:
                for msg in msgs:
                    self.__update_state(msg)

                self.__notify_changes()

    def stop(self) -> None:
        """
            Stop monitoring for state updates.
        """

        self._run = False

    def _my_on_assign(self, consumer, partitions) -> None:

        for p in partitions:
            p.offset = OFFSET_BEGINNING
            self._low, self._high = consumer.get_watermark_offsets(p)

            if self._high == 0:
                self._end_reached = True

        consumer.assign(partitions)



# Copyright 2020 Confluent Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
#
# See: https://github.com/confluentinc/confluent-kafka-python/pull/1088
#

from json import loads
from struct import unpack, pack

from confluent_kafka.schema_registry import topic_subject_name_strategy, _MAGIC_BYTE
from confluent_kafka.schema_registry.avro import AvroDeserializer, _ContextStringIO, _schema_loads, AvroSerializer
from confluent_kafka.serialization import SerializationError
from fastavro import parse_schema, schemaless_reader, schemaless_writer


class AvroSerializerWithReferences(AvroSerializer):
    __slots__ = ['_hash', '_auto_register', '_known_subjects', '_parsed_schema',
                 '_registry', '_schema', '_schema_id', '_schema_name',
                 '_subject_name_func', '_to_dict', '_named_schemas']

    _default_conf = {'auto.register.schemas': True,
                     'subject.name.strategy': topic_subject_name_strategy}

    def __init__(self, schema_registry_client, schema,
                 to_dict=None, conf=None, named_schemas={}):
        self._registry = schema_registry_client
        self._schema_id = None
        self._known_subjects = set()

        if to_dict is not None and not callable(to_dict):
            raise ValueError("to_dict must be callable with the signature"
                             " to_dict(object, SerializationContext)->dict")

        self._to_dict = to_dict
        self._named_schemas = named_schemas

        conf_copy = self._default_conf.copy()
        if conf is not None:
            conf_copy.update(conf)

        self._auto_register = conf_copy.pop('auto.register.schemas')
        if not isinstance(self._auto_register, bool):
            raise ValueError("auto.register.schemas must be a boolean value")

        self._subject_name_func = conf_copy.pop('subject.name.strategy')
        if not callable(self._subject_name_func):
            raise ValueError("subject.name.strategy must be callable")

        schema_dict = loads(schema.schema_str)
        parsed_schema = parse_schema(schema_dict, named_schemas=self._named_schemas)
        schema_name = parsed_schema.get('name', schema_dict['type'])

        self._schema = schema
        self._schema_name = schema_name
        self._parsed_schema = parsed_schema

    def __call__(self, obj, ctx):
        if obj is None:
            return None

        subject = self._subject_name_func(ctx, self._schema_name)

        if self._auto_register and subject not in self._known_subjects:
            self._schema_id = self._registry.register_schema(subject,
                                                             self._schema)
            self._known_subjects.add(subject)
        elif not self._auto_register and subject not in self._known_subjects:
            registered_schema = self._registry.lookup_schema(subject,
                                                             self._schema)
            self._schema_id = registered_schema.schema_id
            self._known_subjects.add(subject)

        if self._to_dict is not None:
            value = self._to_dict(obj, ctx)
        else:
            value = obj

        with _ContextStringIO() as fo:
            fo.write(pack('>bI', _MAGIC_BYTE, self._schema_id))
            schemaless_writer(fo, self._parsed_schema, value)

            return fo.getvalue()


class AvroDeserializerWithReferences(AvroDeserializer):
    __slots__ = ['_reader_schema', '_registry', '_from_dict', '_writer_schemas', '_return_record_name',
                 '_named_schemas']

    def __init__(self, schema_registry_client, schema=None, from_dict=None, return_record_name=False, named_schemas={}):
        self._registry = schema_registry_client
        self._writer_schemas = {}

        self._reader_schema = schema
        self._named_schemas = named_schemas

        if from_dict is not None and not callable(from_dict):
            raise ValueError("from_dict must be callable with the signature"
                             " from_dict(SerializationContext, dict) -> object")
        self._from_dict = from_dict

        self._return_record_name = return_record_name
        if not isinstance(self._return_record_name, bool):
            raise ValueError("return_record_name must be a boolean value")

    def __call__(self, value, ctx):
        if value is None:
            return None

        if len(value) <= 5:
            raise SerializationError("Message too small. This message was not"
                                     " produced with a Confluent"
                                     " Schema Registry serializer")

        with _ContextStringIO(value) as payload:
            magic, schema_id = unpack('>bI', payload.read(5))
            if magic != _MAGIC_BYTE:
                raise SerializationError("Unknown magic byte. This message was"
                                         " not produced with a Confluent"
                                         " Schema Registry serializer")

            writer_schema = self._writer_schemas.get(schema_id, None)

            if writer_schema is None:
                schema = self._registry.get_schema(schema_id)
                prepared_schema = _schema_loads(schema.schema_str)
                writer_schema = parse_schema(loads(
                    prepared_schema.schema_str), named_schemas=self._named_schemas)
                self._writer_schemas[schema_id] = writer_schema

            obj_dict = schemaless_reader(payload,
                                         writer_schema,
                                         self._reader_schema,
                                         self._return_record_name)

            if self._from_dict is not None:
                return self._from_dict(obj_dict, ctx)

            return obj_dict



from abc import ABC, abstractmethod


class EventSourceListener(ABC):
    @abstractmethod
    def on_highwater(self) -> None:
        pass

    @abstractmethod
    def on_highwater_timeout(self) -> None:
        pass

    @abstractmethod
    def on_batch(self, msgs) -> None:
        pass



"""
    Serialization and Deserialization utilities
"""

import pkgutil
from json import loads

from confluent_kafka.schema_registry import SchemaReference, Schema
from confluent_kafka.schema_registry.avro import AvroSerializer, AvroDeserializer
from fastavro import parse_schema

from jlab_jaws.avro.entities import AlarmLocation, AlarmPriority
from jlab_jaws.avro.entities import SimpleProducer, AlarmInstance, AlarmActivationUnion, SimpleAlarming, \
    EPICSAlarming, NoteAlarming, DisabledOverride, FilteredOverride, LatchedOverride, MaskedOverride, OnDelayedOverride, \
    OffDelayedOverride, ShelvedOverride, AlarmOverrideUnion, OverriddenAlarmType, AlarmOverrideKey, ShelvedReason, \
    EPICSSEVR, EPICSSTAT, UnionEncoding, CALCProducer, EPICSProducer, AlarmClass, EffectiveRegistration, \
    EffectiveActivation, EffectiveAlarm, IntermediateMonolog, AlarmState, AlarmOverrideSet, ProcessorTransitions
from jlab_jaws.serde.avro import AvroDeserializerWithReferences, AvroSerializerWithReferences


def _unwrap_enum(value, enum_class):
    """
    When instantiating classes using _from_dict often a variable intended to be an enum is encountered that
    may actually be a String, a Tuple, or an Enum so this function attempts to convert to an Enum if needed.

    A tuple is allowed due to fastavro supporting tuples for complex types.

    :param value: The value to massage into the correct type
    :param enum_class: Enum class to instantiate
    :return: A value likely as an Enum or None
    """

    if value is None:
        result = None
    elif type(value) is tuple:
        result = enum_class[value[1]]
    elif type(value) is str:
        result = enum_class[value]
    else:  # return as is (hopefully already an Enum)
        result = value
    return result


class AlarmClassSerde:
    """
        Provides AlarmClass serde utilities
    """

    @staticmethod
    def to_dict(obj):
        """
        Converts an AlarmClass to a dict.

        :param obj: The AlarmClass
        :return: A dict
        """

        if obj is None:
            return None

        return {
            "category": obj.category.name,
            "priority": obj.priority.name,
            "rationale": obj.rationale,
            "correctiveaction": obj.corrective_action,
            "pointofcontactusername": obj.point_of_contact_username,
            "latching": obj.latching,
            "filterable": obj.filterable,
            "ondelayseconds": obj.on_delay_seconds,
            "offdelayseconds": obj.off_delay_seconds,
            "screenpath": obj.screen_path,
        }

    @staticmethod
    def _to_dict_with_ctx(obj, ctx):
        return AlarmClassSerde.to_dict(obj)

    @staticmethod
    def from_dict(the_dict):
        """
        Converts a dict to an AlarmClass.

        :param the_dict: The dict
        :return: The AlarmClass
        """
        if the_dict is None:
            return None

        return AlarmClass(the_dict.get('category'),
                          _unwrap_enum(the_dict.get('priority'), AlarmPriority),
                          the_dict.get('rationale'),
                          the_dict.get('correctiveaction'),
                          the_dict.get('pointofcontactusername'),
                          the_dict.get('latching'),
                          the_dict.get('filterable'),
                          the_dict.get('ondelayseconds'),
                          the_dict.get('offdelayseconds'),
                          the_dict.get('screenpath'))

    @staticmethod
    def _from_dict_with_ctx(the_dict, ctx):
        return AlarmClassSerde.from_dict(the_dict)

    @staticmethod
    def references():
        priority_schema_ref = SchemaReference("org.jlab.jaws.entity.AlarmPriority", "alarm-priority", 1)

        return [priority_schema_ref]

    @staticmethod
    def named_schemas():
        priority_bytes = pkgutil.get_data("jlab_jaws", "avro/schemas/AlarmPriority.avsc")
        priority_schema_str = priority_bytes.decode('utf-8')

        named_schemas = {}
        ref_dict = loads(priority_schema_str)
        parse_schema(ref_dict, named_schemas=named_schemas)

        return named_schemas

    @staticmethod
    def deserializer(schema_registry_client):
        """
            Return an AlarmClass deserializer.

            :param schema_registry_client: The Confluent Schema Registry Client
            :return: Deserializer
        """
        return AvroDeserializerWithReferences(schema_registry_client, None,
                                              AlarmClassSerde._from_dict_with_ctx, True,
                                              AlarmClassSerde.named_schemas())

    @staticmethod
    def serializer(schema_registry_client):
        """
            Return an AlarmClass serializer.

            :param schema_registry_client: The Confluent Schema Registry client
            :return: Serializer
        """
        value_bytes = pkgutil.get_data("jlab_jaws", "avro/schemas/AlarmClass.avsc")
        value_schema_str = value_bytes.decode('utf-8')

        schema = Schema(value_schema_str, "AVRO",
                        AlarmClassSerde.references())

        return AvroSerializerWithReferences(schema_registry_client, schema,
                                            AlarmClassSerde._to_dict_with_ctx, None,
                                            AlarmClassSerde.named_schemas())


class AlarmInstanceSerde:
    """
        Provides AlarmInstance serde utilities
    """

    @staticmethod
    def to_dict(obj, union_encoding=UnionEncoding.TUPLE):
        """
        Converts an AlarmInstance to a dict.

        :param obj: The AlarmInstance
        :param union_encoding: How the union should be encoded
        :return: A dict
        """

        if obj is None:
            return None

        if isinstance(obj.producer, SimpleProducer):
            uniontype = "org.jlab.jaws.entity.SimpleProducer"
            uniondict = {}
        elif isinstance(obj.producer, EPICSProducer):
            uniontype = "org.jlab.jaws.entity.EPICSProducer"
            uniondict = {"pv": obj.producer.pv}
        elif isinstance(obj.producer, CALCProducer):
            uniontype = "org.jlab.jaws.entity.CALCProducer"
            uniondict = {"expression": obj.producer.expression}
        else:
            raise Exception("Unknown alarming union type: {}".format(obj.producer))

        if union_encoding is UnionEncoding.TUPLE:
            union = (uniontype, uniondict)
        elif union_encoding is UnionEncoding.DICT_WITH_TYPE:
            union = {uniontype: uniondict}
        else:
            union = uniondict

        return {
            "class": obj.alarm_class,
            "producer": union,
            "location": obj.location,
            "maskedby": obj.masked_by,
            "screenpath": obj.screen_path
        }

    @staticmethod
    def _to_dict_with_ctx(obj, ctx):
        return AlarmInstanceSerde.to_dict(obj)

    @staticmethod
    def from_dict(the_dict):
        """
        Converts a dict to an AlarmInstance.

        Note: UnionEncoding.POSSIBLY_AMBIGUOUS_DICT is not supported.

        :param the_dict: The dict
        :return: The AlarmInstance
        """

        if the_dict is None:
            return None

        unionobj = the_dict['producer']

        if type(unionobj) is tuple:
            uniontype = unionobj[0]
            uniondict = unionobj[1]
        elif type(unionobj is dict):
            value = next(iter(unionobj.items()))
            uniontype = value[0]
            uniondict = value[1]
        else:
            raise Exception("Unsupported union encoding")

        if uniontype == "org.jlab.jaws.entity.CalcProducer":
            producer = CALCProducer(uniondict['expression'])
        elif uniontype == "org.jlab.jaws.entity.EPICSProducer":
            producer = EPICSProducer(uniondict['pv'])
        else:
            producer = SimpleProducer()

        return AlarmInstance(the_dict.get('class'),
                             producer,
                             the_dict.get('location'),
                             the_dict.get('maskedby'),
                             the_dict.get('screenpath'))


    @staticmethod
    def _from_dict_with_ctx(the_dict, ctx):
        return AlarmInstanceSerde.from_dict(the_dict)

    @staticmethod
    def deserializer(schema_registry_client):
        """
            Return an AlarmInstance deserializer.

            :param schema_registry_client: The Confluent Schema Registry Client
            :return: Deserializer
        """
        return AvroDeserializer(schema_registry_client, None,
                                AlarmInstanceSerde._from_dict_with_ctx, True)

    @staticmethod
    def serializer(schema_registry_client):
        """
            Return an AlarmInstance serializer.

            :param schema_registry_client: The Confluent Schema Registry client
            :return: Serializer
        """
        value_bytes = pkgutil.get_data("jlab_jaws", "avro/schemas/AlarmInstance.avsc")
        value_schema_str = value_bytes.decode('utf-8')

        return AvroSerializer(schema_registry_client, value_schema_str,
                              AlarmInstanceSerde._to_dict_with_ctx, None)


class AlarmActivationUnionSerde:
    """
        Provides AlarmActivationUnion serde utilities
    """

    @staticmethod
    def to_dict(obj, union_encoding=UnionEncoding.TUPLE):
        """
        Converts an AlarmActivationUnion to a dict.

        :param obj: The AlarmActivationUnion
        :param union_encoding: How the union should be encoded
        :return: A dict
        """

        if obj is None:
            return None

        if isinstance(obj.msg, SimpleAlarming):
            uniontype = "org.jlab.jaws.entity.SimpleAlarming"
            uniondict = {}
        elif isinstance(obj.msg, EPICSAlarming):
            uniontype = "org.jlab.jaws.entity.EPICSAlarming"
            uniondict = {"sevr": obj.msg.sevr.name, "stat": obj.msg.stat.name}
        elif isinstance(obj.msg, NoteAlarming):
            uniontype = "org.jlab.jaws.entity.NoteAlarming"
            uniondict = {"note": obj.msg.note}
        else:
            raise Exception("Unknown alarming union type: {}".format(obj.msg))

        if union_encoding is UnionEncoding.TUPLE:
            union = (uniontype, uniondict)
        elif union_encoding is UnionEncoding.DICT_WITH_TYPE:
            union = {uniontype: uniondict}
        else:
            union = uniondict

        return {
            "msg": union
        }

    @staticmethod
    def _to_dict_with_ctx(obj, ctx):
        return AlarmActivationUnionSerde.to_dict(obj)

    @staticmethod
    def from_dict(the_dict):
        """
        Converts a dict to an AlarmActivationUnion.

        Note: UnionEncoding.POSSIBLY_AMBIGUOUS_DICT is not supported.

        :param the_dict: The dict
        :return: The AlarmActivationUnion
        """

        if the_dict is None:
            return None

        unionobj = the_dict['msg']

        if type(unionobj) is tuple:
            uniontype = unionobj[0]
            uniondict = unionobj[1]
        elif type(unionobj is dict):
            value = next(iter(unionobj.items()))
            uniontype = value[0]
            uniondict = value[1]
        else:
            raise Exception("Unsupported union encoding")

        if uniontype == "org.jlab.jaws.entity.NoteAlarming":
            obj = NoteAlarming(uniondict['note'])
        elif uniontype == "org.jlab.jaws.entity.EPICSAlarming":
            obj = EPICSAlarming(_unwrap_enum(uniondict['sevr'], EPICSSEVR), _unwrap_enum(uniondict['stat'],
                                                                                         EPICSSTAT))
        else:
            obj = SimpleAlarming()

        return AlarmActivationUnion(obj)

    @staticmethod
    def _from_dict_with_ctx(the_dict, ctx):
        return AlarmActivationUnionSerde.from_dict(the_dict)

    @staticmethod
    def deserializer(schema_registry_client):
        """
            Return an AlarmActivationUnion deserializer.

            :param schema_registry_client: The Confluent Schema Registry Client
            :return: Deserializer
        """

        return AvroDeserializer(schema_registry_client, None,
                                AlarmActivationUnionSerde._from_dict_with_ctx, True)

    @staticmethod
    def serializer(schema_registry_client):
        """
            Return an AlarmActivationUnion serializer.

            :param schema_registry_client: The Confluent Schema Registry client
            :return: Serializer
        """

        value_bytes = pkgutil.get_data("jlab_jaws", "avro/schemas/AlarmActivationUnion.avsc")
        value_schema_str = value_bytes.decode('utf-8')

        return AvroSerializer(schema_registry_client, value_schema_str,
                              AlarmActivationUnionSerde._to_dict_with_ctx, None)


class AlarmOverrideKeySerde:
    """
        Provides AlarmOverrideKey serde utilities
    """

    @staticmethod
    def to_dict(obj):
        """
        Converts an AlarmOverrideKey to a dict.

        :param obj: The AlarmOverrideKey
        :return: A dict
        """
        return {
            "name": obj.name,
            "type": obj.type.name
        }

    @staticmethod
    def _to_dict_with_ctx(obj, ctx):
        return AlarmOverrideKeySerde.to_dict(obj)

    @staticmethod
    def from_dict(the_dict):
        """
        Converts a dict to an AlarmOverrideKey.

        :param the_dict: The dict
        :return: The AlarmOverrideKey
        """
        return AlarmOverrideKey(the_dict['name'], _unwrap_enum(the_dict['type'], OverriddenAlarmType))

    @staticmethod
    def _from_dict_with_ctx(the_dict, ctx):
        return AlarmOverrideKeySerde.from_dict(the_dict)

    @staticmethod
    def deserializer(schema_registry_client):
        """
            Return an AlarmOverrideKey deserializer.

            :param schema_registry_client: The Confluent Schema Registry Client
            :return: Deserializer
        """

        return AvroDeserializer(schema_registry_client, None,
                                AlarmOverrideKeySerde._from_dict_with_ctx, True)

    @staticmethod
    def serializer(schema_registry_client):
        """
            Return an AlarmOverrideKey serializer.

            :param schema_registry_client: The Confluent Schema Registry client
            :return: Serializer
        """

        subject_bytes = pkgutil.get_data("jlab_jaws", "avro/schemas/AlarmOverrideKey.avsc")
        subject_schema_str = subject_bytes.decode('utf-8')

        return AvroSerializer(schema_registry_client, subject_schema_str,
                              AlarmOverrideKeySerde._to_dict_with_ctx, None)


class AlarmOverrideUnionSerde:
    """
        Provides AlarmOverrideUnion serde utilities
    """

    @staticmethod
    def to_dict(obj, union_encoding=UnionEncoding.TUPLE):
        """
        Converts an AlarmOverrideUnion to a dict.

        :param obj: The AlarmOverrideUnion
        :param union_encoding: How the union should be encoded
        :return: A dict
        """
        if isinstance(obj.msg, DisabledOverride):
            uniontype = "org.jlab.jaws.entity.DisabledOverride"
            uniondict = {"comments": obj.msg.comments}
        elif isinstance(obj.msg, FilteredOverride):
            uniontype = "org.jlab.jaws.entity.FilteredOverride"
            uniondict = {"filtername": obj.msg.filtername}
        elif isinstance(obj.msg, LatchedOverride):
            uniontype = "org.jlab.jaws.entity.LatchedOverride"
            uniondict = {}
        elif isinstance(obj.msg, MaskedOverride):
            uniontype = "org.jlab.jaws.entity.MaskedOverride"
            uniondict = {}
        elif isinstance(obj.msg, OnDelayedOverride):
            uniontype = "org.jlab.jaws.entity.OnDelayedOverride"
            uniondict = {"expiration": obj.msg.expiration}
        elif isinstance(obj.msg, OffDelayedOverride):
            uniontype = "org.jlab.jaws.entity.OffDelayedOverride"
            uniondict = {"expiration": obj.msg.expiration}
        elif isinstance(obj.msg, ShelvedOverride):
            uniontype = "org.jlab.jaws.entity.ShelvedOverride"
            uniondict = {"expiration": obj.msg.expiration, "comments": obj.msg.comments,
                         "reason": obj.msg.reason.name, "oneshot": obj.msg.oneshot}
        else:
            print("Unknown alarming union type: {}".format(obj.msg))
            uniontype = None
            uniondict = None

        if union_encoding is UnionEncoding.TUPLE:
            union = (uniontype, uniondict)
        elif union_encoding is UnionEncoding.DICT_WITH_TYPE:
            union = {uniontype: uniondict}
        else:
            union = uniondict

        return {
            "msg": union
        }

    @staticmethod
    def _to_dict_with_ctx(obj, ctx):
        return AlarmOverrideUnionSerde.to_dict(obj)

    @staticmethod
    def from_dict(the_dict):
        """
        Converts a dict to an AlarmOverrideUnion.

        Note: Both UnionEncoding.TUPLE and UnionEncoding.DICT_WITH_TYPE are supported,
        but UnionEncoding.POSSIBLY_AMBIGUOUS_DICT is not supported at this time
        because I'm lazy and not going to try to guess what type is in your union.

        :param the_dict: The dict (or maybe it's a duck)
        :return: The AlarmOverrideUnion
        """
        alarmingobj = the_dict['msg']

        if type(alarmingobj) is tuple:
            alarmingtype = alarmingobj[0]
            alarmingdict = alarmingobj[1]
        elif type(alarmingobj is dict):
            value = next(iter(alarmingobj.items()))
            alarmingtype = value[0]
            alarmingdict = value[1]
        else:
            raise Exception("Unsupported union encoding")

        if alarmingtype == "org.jlab.jaws.entity.DisabledOverride":
            obj = DisabledOverride(alarmingdict['comments'])
        elif alarmingtype == "org.jlab.jaws.entity.FilteredOverride":
            obj = FilteredOverride(alarmingdict['filtername'])
        elif alarmingtype == "org.jlab.jaws.entity.LatchedOverride":
            obj = LatchedOverride()
        elif alarmingtype == "org.jlab.jaws.entity.MaskedOverride":
            obj = MaskedOverride()
        elif alarmingtype == "org.jlab.jaws.entity.OnDelayedOverride":
            obj = OnDelayedOverride(alarmingdict['expiration'])
        elif alarmingtype == "org.jlab.jaws.entity.OffDelayedOverride":
            obj = OffDelayedOverride(alarmingdict['expiration'])
        elif alarmingtype == "org.jlab.jaws.entity.ShelvedOverride":
            obj = ShelvedOverride(alarmingdict['expiration'], alarmingdict['comments'],
                                  _unwrap_enum(alarmingdict['reason'], ShelvedReason), alarmingdict['oneshot'])
        else:
            print("Unknown alarming type: {}".format(the_dict['msg']))
            obj = None

        return AlarmOverrideUnion(obj)

    @staticmethod
    def _from_dict_with_ctx(the_dict, ctx):
        return AlarmOverrideUnionSerde.from_dict(the_dict)

    @staticmethod
    def references():
        disabled_schema_ref = SchemaReference("org.jlab.jaws.entity.DisabledOverride", "disabled-override", 1)
        filtered_schema_ref = SchemaReference("org.jlab.jaws.entity.FilteredOverride", "filtered-override", 1)
        latched_schema_ref = SchemaReference("org.jlab.jaws.entity.LatchedOverride", "latched-override", 1)
        masked_schema_ref = SchemaReference("org.jlab.jaws.entity.MaskedOverride", "masked-override", 1)
        off_delayed_schema_ref = SchemaReference("org.jlab.jaws.entity.OffDelayedOverride", "off-delayed-override", 1)
        on_delayed_schema_ref = SchemaReference("org.jlab.jaws.entity.OnDelayedOverride", "on-delayed-override", 1)
        shelved_schema_ref = SchemaReference("org.jlab.jaws.entity.ShelvedOverride", "shelved-override", 1)

        return [disabled_schema_ref,
                filtered_schema_ref,
                latched_schema_ref,
                masked_schema_ref,
                off_delayed_schema_ref,
                on_delayed_schema_ref,
                shelved_schema_ref]

    @staticmethod
    def named_schemas():
        disabled_bytes = pkgutil.get_data("jlab_jaws", "avro/schemas/DisabledOverride.avsc")
        disabled_schema_str = disabled_bytes.decode('utf-8')

        filtered_bytes = pkgutil.get_data("jlab_jaws", "avro/schemas/FilteredOverride.avsc")
        filtered_schema_str = filtered_bytes.decode('utf-8')

        latched_bytes = pkgutil.get_data("jlab_jaws", "avro/schemas/LatchedOverride.avsc")
        latched_schema_str = latched_bytes.decode('utf-8')

        masked_bytes = pkgutil.get_data("jlab_jaws", "avro/schemas/MaskedOverride.avsc")
        masked_schema_str = masked_bytes.decode('utf-8')

        off_delayed_bytes = pkgutil.get_data("jlab_jaws", "avro/schemas/OffDelayedOverride.avsc")
        off_delayed_schema_str = off_delayed_bytes.decode('utf-8')

        on_delayed_bytes = pkgutil.get_data("jlab_jaws", "avro/schemas/OnDelayedOverride.avsc")
        on_delayed_schema_str = on_delayed_bytes.decode('utf-8')

        shelved_bytes = pkgutil.get_data("jlab_jaws", "avro/schemas/ShelvedOverride.avsc")
        shelved_schema_str = shelved_bytes.decode('utf-8')

        named_schemas = {}
        ref_dict = loads(disabled_schema_str)
        parse_schema(ref_dict, named_schemas=named_schemas)
        ref_dict = loads(filtered_schema_str)
        parse_schema(ref_dict, named_schemas=named_schemas)
        ref_dict = loads(latched_schema_str)
        parse_schema(ref_dict, named_schemas=named_schemas)
        ref_dict = loads(masked_schema_str)
        parse_schema(ref_dict, named_schemas=named_schemas)
        ref_dict = loads(off_delayed_schema_str)
        parse_schema(ref_dict, named_schemas=named_schemas)
        ref_dict = loads(on_delayed_schema_str)
        parse_schema(ref_dict, named_schemas=named_schemas)
        ref_dict = loads(shelved_schema_str)
        parse_schema(ref_dict, named_schemas=named_schemas)

        return named_schemas

    @staticmethod
    def deserializer(schema_registry_client):
        """
            Return an AlarmOverrideUnion deserializer.

            :param schema_registry_client: The Confluent Schema Registry Client
            :return: Deserializer
        """

        return AvroDeserializerWithReferences(schema_registry_client, None,
                                              AlarmOverrideUnionSerde._from_dict_with_ctx, True,
                                              AlarmOverrideUnionSerde.named_schemas())

    @staticmethod
    def serializer(schema_registry_client):
        """
            Return an AlarmOverrideUnion serializer.

            :param schema_registry_client: The Confluent Schema Registry client
            :return: Serializer
        """

        subject_bytes = pkgutil.get_data("jlab_jaws", "avro/schemas/AlarmOverrideUnion.avsc")
        subject_schema_str = subject_bytes.decode('utf-8')

        schema = Schema(subject_schema_str, "AVRO",
                        AlarmOverrideUnionSerde.references())

        return AvroSerializerWithReferences(schema_registry_client, schema,
                                            AlarmOverrideUnionSerde._to_dict_with_ctx, None,
                                            AlarmOverrideUnionSerde.named_schemas())


class AlarmOverrideSetSerde:
    """
        Provides AlarmOverrideSet serde utilities
    """

    @staticmethod
    def to_dict(obj):
        """
        Converts AlarmOverrideSet to a dict

        :param obj: The AlarmOverrideSet
        :return: A dict
        """
        return {
            "disabled": {"comments": obj.disabled.comments} if obj.disabled is not None else None,
            "filtered": {"filtername": obj.filtered.filtername} if obj.filtered is not None else None,
            "latched": {} if obj.latched is not None else None,
            "masked": {} if obj.masked is not None else None,
            "ondelayed": {"expiration": obj.ondelayed.expiration} if obj.ondelayed is not None else None,
            "offdelayed": {"expiration": obj.offdelayed.expiration} if obj.offdelayed is not None else None,
            "shelved": {"expiration": obj.shelved.expiration,
                        "comments": obj.shelved.comments,
                        "oneshot": obj.shelved.oneshot,
                        "reason": obj.shelved.reason.name} if obj.shelved is not None else None
        }

    @staticmethod
    def from_dict(the_dict):
        """
        Converts a dict to AlarmOverrideSet.

        :param the_dict: The dict
        :return: The AlarmOverrideSet
        """
        return AlarmOverrideSet(DisabledOverride(the_dict['disabled'][1]['comments'])
                                if the_dict.get('disabled') is not None else None,
                                FilteredOverride(the_dict['filtered'][1]['filtername'])
                                if the_dict.get('filtered') is not None else None,
                                LatchedOverride()
                                if the_dict.get('latched') is not None else None,
                                MaskedOverride()
                                if the_dict.get('masked') is not None else None,
                                OnDelayedOverride(the_dict['ondelayed'][1]['expiration'])
                                if the_dict.get('ondelayed') is not None else None,
                                OffDelayedOverride(the_dict['offdelayed'][1]['expiration'])
                                if the_dict.get('offdelayed') is not None else None,
                                ShelvedOverride(the_dict['shelved'][1]['expiration'],
                                                the_dict['shelved'][1]['comments'],
                                                ShelvedReason[the_dict['shelved'][1]['reason']],
                                                the_dict['shelved'][1]['oneshot'])
                                if the_dict.get('shelved') is not None else None)


class ProcessorTransitionsSerde:
    """
        Provides ProcessorTransitions serde utilities
    """

    @staticmethod
    def to_dict(obj):
        """
        Converts ProcessorTransitions to a dict

        :param obj: The ProcessorTransitions
        :return: A dict
        """
        return {
            "transitionToActive": obj.transitionToActive,
            "transitionToNormal": obj.transitionToNormal,
            "latching": obj.latching,
            "unshelving": obj.unshelving,
            "masking": obj.masking,
            "unmasking": obj.unmasking,
            "ondelaying": obj.ondelaying,
            "offdelaying": obj.offdelaying
        }

    @staticmethod
    def from_dict(the_dict):
        """
        Converts a dict to ProcessorTransitions.

        :param the_dict: The dict
        :return: The ProcessorTransitions
        """
        return ProcessorTransitions(the_dict['transitionToActive'],
                                    the_dict['transitionToNormal'],
                                    the_dict['latching'],
                                    the_dict['unshelving'],
                                    the_dict['masking'],
                                    the_dict['unmasking'],
                                    the_dict['ondelaying'],
                                    the_dict['offdelaying'])


class EffectiveRegistrationSerde:
    """
        Provides EffectiveRegistration serde utilities
    """

    @staticmethod
    def to_dict(obj):
        """
        Converts EffectiveRegistration to a dict.

        :param obj: The EffectiveRegistration
        :return: A dict
        """
        return {
            "class": AlarmClassSerde.to_dict(obj.alarm_class),
            "actual": AlarmInstanceSerde.to_dict(obj.actual),
            "calculated": AlarmInstanceSerde.to_dict(obj.calculated)
        }

    @staticmethod
    def _to_dict_with_ctx(obj, ctx):
        return EffectiveRegistrationSerde.to_dict(obj)

    @staticmethod
    def from_dict(the_dict):
        """
        Converts a dict to EffectiveRegistration.

        :param the_dict: The dict
        :return: The EffectiveRegistration
        """
        return EffectiveRegistration(
            AlarmClassSerde.from_dict(the_dict['class'][1]) if the_dict.get('class') is not None else None,
            AlarmInstanceSerde.from_dict(the_dict['actual'][1])
            if the_dict.get('actual') is not None else None,
            AlarmInstanceSerde.from_dict(the_dict['calculated'][1])
            if the_dict.get('calculated') is not None else None)

    @staticmethod
    def _from_dict_with_ctx(the_dict, ctx):
        return EffectiveRegistrationSerde.from_dict(the_dict)

    @staticmethod
    def references():
        references = []

        classes_schema_ref = SchemaReference("org.jlab.jaws.entity.AlarmClass", "alarm-classes-value", 1)
        registration_schema_ref = SchemaReference("org.jlab.jaws.entity.AlarmInstance",
                                                  "alarm-instances-value", 1)

        references.append(classes_schema_ref)
        references.append(registration_schema_ref)

        return references

    @staticmethod
    def named_schemas():
        classes_bytes = pkgutil.get_data("jlab_jaws", "avro/schemas/AlarmClass.avsc")
        classes_schema_str = classes_bytes.decode('utf-8')

        registrations_bytes = pkgutil.get_data("jlab_jaws", "avro/schemas/AlarmInstance.avsc")
        registrations_schema_str = registrations_bytes.decode('utf-8')

        named_schemas = AlarmClassSerde.named_schemas()

        named_schemas.update(AlarmOverrideUnionSerde.named_schemas())

        ref_dict = loads(classes_schema_str)
        parse_schema(ref_dict, named_schemas=named_schemas)
        ref_dict = loads(registrations_schema_str)
        parse_schema(ref_dict, named_schemas=named_schemas)

        return named_schemas

    @staticmethod
    def deserializer(schema_registry_client):
        """
            Return an EffectiveRegistration deserializer.

            :param schema_registry_client: The Confluent Schema Registry Client
            :return: Deserializer
        """

        return AvroDeserializerWithReferences(schema_registry_client,
                                              None,
                                              EffectiveRegistrationSerde._from_dict_with_ctx,
                                              True,
                                              EffectiveRegistrationSerde.named_schemas())

    @staticmethod
    def serializer(schema_registry_client, conf=None):
        """
            Return an EffectiveRegistration serializer.

            :param conf: Configuration
            :param schema_registry_client: The Confluent Schema Registry client
            :return: Serializer
        """

        subject_bytes = pkgutil.get_data("jlab_jaws", "avro/schemas/EffectiveRegistration.avsc")
        subject_schema_str = subject_bytes.decode('utf-8')

        schema = Schema(subject_schema_str, "AVRO",
                        EffectiveRegistrationSerde.references())

        return AvroSerializerWithReferences(schema_registry_client,
                                            schema,
                                            EffectiveRegistrationSerde._to_dict_with_ctx,
                                            conf,
                                            EffectiveRegistrationSerde.named_schemas())


class EffectiveActivationSerde:
    """
        Provides EffectiveActivation serde utilities
    """

    @staticmethod
    def to_dict(obj):
        """
        Converts EffectiveActivation to a dict.

        :param obj: The EffectiveActivation
        :return: A dict
        """
        return {
            "actual": AlarmActivationUnionSerde.to_dict(obj.actual),
            "overrides": AlarmOverrideSetSerde.to_dict(obj.overrides),
            "state": obj.state.name
        }

    @staticmethod
    def _to_dict_with_ctx(obj, ctx):
        return EffectiveActivationSerde.to_dict(obj)

    @staticmethod
    def from_dict(the_dict):
        """
        Converts a dict to EffectiveActivation.

        :param the_dict: The dict
        :return: The EffectiveActivation
        """
        return EffectiveActivation(
            AlarmActivationUnionSerde.from_dict(the_dict['actual'][1])
            if the_dict.get('actual') is not None else None,
            AlarmOverrideSetSerde.from_dict(the_dict['overrides']),
            _unwrap_enum(the_dict['state'], AlarmState))

    @staticmethod
    def _from_dict_with_ctx(the_dict, ctx):
        return EffectiveActivationSerde.from_dict(the_dict)

    @staticmethod
    def references():
        references = []

        activation_schema_ref = SchemaReference("org.jlab.jaws.entity.AlarmActivationUnion",
                                                "alarm-activations-value", 1)
        overrides_schema_ref = SchemaReference("org.jlab.jaws.entity.AlarmOverrideSet",
                                               "alarm-override-set", 1)
        state_schema_ref = SchemaReference("org.jlab.jaws.entity.AlarmState", "alarm-state", 1)

        references.append(activation_schema_ref)
        references.append(overrides_schema_ref)
        references.append(state_schema_ref)

        return references

    @staticmethod
    def named_schemas():
        activation_bytes = pkgutil.get_data("jlab_jaws", "avro/schemas/AlarmActivationUnion.avsc")
        activation_schema_str = activation_bytes.decode('utf-8')

        overrides_bytes = pkgutil.get_data("jlab_jaws", "avro/schemas/AlarmOverrideSet.avsc")
        overrides_schema_str = overrides_bytes.decode('utf-8')

        state_bytes = pkgutil.get_data("jlab_jaws", "avro/schemas/AlarmState.avsc")
        state_schema_str = state_bytes.decode('utf-8')

        named_schemas = AlarmClassSerde.named_schemas()

        named_schemas.update(AlarmOverrideUnionSerde.named_schemas())

        ref_dict = loads(activation_schema_str)
        parse_schema(ref_dict, named_schemas=named_schemas)
        ref_dict = loads(overrides_schema_str)
        parse_schema(ref_dict, named_schemas=named_schemas)
        ref_dict = loads(state_schema_str)
        parse_schema(ref_dict, named_schemas=named_schemas)

        return named_schemas

    @staticmethod
    def deserializer(schema_registry_client):
        """
            Return an EffectiveActivation deserializer.

            :param schema_registry_client: The Confluent Schema Registry Client
            :return: Deserializer
        """

        return AvroDeserializerWithReferences(schema_registry_client,
                                              None,
                                              EffectiveActivationSerde._from_dict_with_ctx,
                                              True,
                                              EffectiveActivationSerde.named_schemas())

    @staticmethod
    def serializer(schema_registry_client, conf=None):
        """
            Return an EffectiveActivation serializer.

            :param conf: Configuration
            :param schema_registry_client: The Confluent Schema Registry client
            :return: Serializer
        """

        subject_bytes = pkgutil.get_data("jlab_jaws", "avro/schemas/EffectiveActivation.avsc")
        subject_schema_str = subject_bytes.decode('utf-8')

        schema = Schema(subject_schema_str, "AVRO",
                        EffectiveActivationSerde.references())

        return AvroSerializerWithReferences(schema_registry_client,
                                            schema,
                                            EffectiveActivationSerde._to_dict_with_ctx,
                                            conf,
                                            EffectiveActivationSerde.named_schemas())


class EffectiveAlarmSerde:
    """
        Provides EffectiveAlarm serde utilities
    """

    @staticmethod
    def to_dict(obj):
        """
        Converts EffectiveAlarm to a dict.

        :param obj: The EffectiveAlarm
        :return: A dict
        """
        return {
            "registration": EffectiveRegistrationSerde.to_dict(obj.registration),
            "activation": EffectiveActivationSerde.to_dict(obj.activation),
        }

    @staticmethod
    def _to_dict_with_ctx(obj, ctx):
        return EffectiveAlarmSerde.to_dict(obj)

    @staticmethod
    def from_dict(the_dict):
        """
        Converts a dict to EffectiveAlarm.

        :param the_dict: The dict
        :return: The EffectiveAlarm
        """
        return EffectiveAlarm(EffectiveRegistrationSerde.from_dict(the_dict['registration']),
                              EffectiveActivationSerde.from_dict(the_dict['activation']))

    @staticmethod
    def _from_dict_with_ctx(the_dict, ctx):
        return EffectiveAlarmSerde.from_dict(the_dict)

    @staticmethod
    def references():
        references = []

        registration_schema_ref = SchemaReference("org.jlab.jaws.entity.EffectiveRegistration",
                                                  "effective-registrations-value", 1)
        activation_schema_ref = SchemaReference("org.jlab.jaws.entity.EffectiveActivation",
                                                "effective-activations-value", 1)

        references.append(registration_schema_ref)
        references.append(activation_schema_ref)

        return references

    @staticmethod
    def named_schemas():
        registrations_bytes = pkgutil.get_data("jlab_jaws", "avro/schemas/EffectiveRegistration.avsc")
        registrations_schema_str = registrations_bytes.decode('utf-8')

        activation_bytes = pkgutil.get_data("jlab_jaws", "avro/schemas/EffectiveActivation.avsc")
        activation_schema_str = activation_bytes.decode('utf-8')

        named_schemas = EffectiveRegistrationSerde.named_schemas()
        named_schemas.update(EffectiveActivationSerde.named_schemas())

        ref_dict = loads(registrations_schema_str)
        parse_schema(ref_dict, named_schemas=named_schemas)
        ref_dict = loads(activation_schema_str)
        parse_schema(ref_dict, named_schemas=named_schemas)

        return named_schemas

    @staticmethod
    def deserializer(schema_registry_client):
        """
            Return an EffectiveAlarm deserializer.

            :param schema_registry_client: The Confluent Schema Registry Client
            :return: Deserializer
        """

        return AvroDeserializerWithReferences(schema_registry_client,
                                              None,
                                              EffectiveAlarmSerde._from_dict_with_ctx,
                                              True,
                                              EffectiveAlarmSerde.named_schemas())

    @staticmethod
    def serializer(schema_registry_client, conf=None):
        """
            Return an EffectiveAlarm serializer.

            :param conf: Configuration
            :param schema_registry_client: The Confluent Schema Registry client
            :return: Serializer
        """

        subject_bytes = pkgutil.get_data("jlab_jaws", "avro/schemas/EffectiveAlarm.avsc")
        subject_schema_str = subject_bytes.decode('utf-8')

        schema = Schema(subject_schema_str, "AVRO",
                        EffectiveAlarmSerde.references())

        return AvroSerializerWithReferences(schema_registry_client,
                                            schema,
                                            EffectiveAlarmSerde._to_dict_with_ctx,
                                            conf,
                                            EffectiveAlarmSerde.named_schemas())


class IntermediateMonologSerde:
    """
        Provides IntermediateMonolog serde utilities
    """

    @staticmethod
    def to_dict(obj):
        """
        Converts IntermediateMonolog to a dict.

        :param obj: The IntermediateMonolog
        :return: A dict
        """
        return {
            "registration": EffectiveRegistrationSerde.to_dict(obj.registration),
            "activation": EffectiveActivationSerde.to_dict(obj.activation),
            "transitions": ProcessorTransitionsSerde.to_dict(obj.transitions)
        }

    @staticmethod
    def _to_dict_with_ctx(obj, ctx):
        return EffectiveAlarmSerde.to_dict(obj)

    @staticmethod
    def from_dict(the_dict):
        """
        Converts a dict to IntermediateMonolog.

        :param the_dict: The dict
        :return: The IntermediateMonolog
        """
        return IntermediateMonolog(EffectiveRegistrationSerde.from_dict(the_dict['registration']),
                                   EffectiveActivationSerde.from_dict(the_dict['activation']),
                                   ProcessorTransitionsSerde.from_dict(the_dict['transitions']))

    @staticmethod
    def _from_dict_with_ctx(the_dict, ctx):
        return IntermediateMonologSerde.from_dict(the_dict)

    @staticmethod
    def references():
        references = []

        registration_schema_ref = SchemaReference("org.jlab.jaws.entity.EffectiveRegistration",
                                                  "effective-registrations-value", 1)

        activation_schema_ref = SchemaReference("org.jlab.jaws.entity.EffectiveActivation",
                                                "effective-activations-value", 1)

        transitions_schema_ref = SchemaReference("org.jlab.jaws.entity.ProcessorTransitions",
                                                 "processor-transitions", 1)

        references.append(registration_schema_ref)
        references.append(activation_schema_ref)
        references.append(transitions_schema_ref)

        return references

    @staticmethod
    def named_schemas():
        registrations_bytes = pkgutil.get_data("jlab_jaws", "avro/schemas/EffectiveRegistration.avsc")
        registrations_schema_str = registrations_bytes.decode('utf-8')

        activation_bytes = pkgutil.get_data("jlab_jaws", "avro/schemas/EffectiveActivation.avsc")
        activation_schema_str = activation_bytes.decode('utf-8')

        transitions_bytes = pkgutil.get_data("jlab_jaws", "avro/schemas/ProcessorTransitions.avsc")
        transitions_schema_str = transitions_bytes.decode('utf-8')

        named_schemas = EffectiveRegistrationSerde.named_schemas()
        named_schemas.update(EffectiveActivationSerde.named_schemas())

        ref_dict = loads(registrations_schema_str)
        parse_schema(ref_dict, named_schemas=named_schemas)

        ref_dict = loads(activation_schema_str)
        parse_schema(ref_dict, named_schemas=named_schemas)

        ref_dict = loads(transitions_schema_str)
        parse_schema(ref_dict, named_schemas=named_schemas)

        return named_schemas

    @staticmethod
    def deserializer(schema_registry_client):
        """
            Return an IntermediateMonolog deserializer.

            :param schema_registry_client: The Confluent Schema Registry Client
            :return: Deserializer
        """

        return AvroDeserializerWithReferences(schema_registry_client,
                                              None,
                                              IntermediateMonologSerde._from_dict_with_ctx,
                                              True,
                                              IntermediateMonologSerde.named_schemas())

    @staticmethod
    def serializer(schema_registry_client, conf=None):
        """
            Return an IntermediateMonolog serializer.

            :param conf: Configuration
            :param schema_registry_client: The Confluent Schema Registry client
            :return: Serializer
        """

        subject_bytes = pkgutil.get_data("jlab_jaws", "avro/schemas/IntermediateMonolog.avsc")
        subject_schema_str = subject_bytes.decode('utf-8')

        schema = Schema(subject_schema_str, "AVRO",
                        IntermediateMonologSerde.references())

        return AvroSerializerWithReferences(schema_registry_client,
                                            schema,
                                            IntermediateMonologSerde._to_dict_with_ctx,
                                            conf,
                                            IntermediateMonologSerde.named_schemas())


class AlarmLocationSerde:
    """
        Provides AlarmLocation serde utilities
    """

    @staticmethod
    def to_dict(obj):
        """
        Converts AlarmLocation to a dict.

        :param obj: The AlarmLocation
        :return: A dict
        """
        return {
            "parent": obj.parent
        }

    @staticmethod
    def _to_dict_with_ctx(obj, ctx):
        return AlarmLocationSerde.to_dict(obj)

    @staticmethod
    def from_dict(the_dict):
        """
        Converts a dict to AlarmLocation.

        :param the_dict: The dict
        :return: The AlarmLocation
        """
        return AlarmLocation(the_dict['parent'])

    @staticmethod
    def _from_dict_with_ctx(the_dict, ctx):
        return AlarmLocationSerde.from_dict(the_dict)

    @staticmethod
    def references():
        references = []

        return references

    @staticmethod
    def named_schemas():
        named_schemas = {}

        return named_schemas

    @staticmethod
    def deserializer(schema_registry_client):
        """
            Return an AlarmLocation deserializer.

            :param schema_registry_client: The Confluent Schema Registry Client
            :return: Deserializer
        """

        return AvroDeserializer(schema_registry_client,
                                None,
                                AlarmLocationSerde._from_dict_with_ctx,
                                True)

    @staticmethod
    def serializer(schema_registry_client, conf=None):
        """
            Return an AlarmLocation serializer.

            :param conf: Configuration
            :param schema_registry_client: The Confluent Schema Registry client
            :return: Serializer
        """

        subject_bytes = pkgutil.get_data("jlab_jaws", "avro/schemas/AlarmLocation.avsc")
        subject_schema_str = subject_bytes.decode('utf-8')

        return AvroSerializer(schema_registry_client,
                              subject_schema_str,
                              AlarmLocationSerde._to_dict_with_ctx,
                              conf)



"""
    Python entities corresponding to AVRO schemas.
"""

from dataclasses import dataclass
from enum import Enum
from typing import Union, Optional, List


class AlarmPriority(Enum):
    P1_CRITICAL = 1
    P2_MAJOR = 2
    P3_MINOR = 3
    P4_INCIDENTAL = 4


class UnionEncoding(Enum):
    """
        Enum of possible ways to encode an AVRO union instance in Python.
    """
    TUPLE = 1
    """Instance is a two-tuple of (str - type name, dict); this is how fastavro works 
    (https://fastavro.readthedocs.io/en/latest/writer.html#using-the-tuple-notation-to-specify-which-branch-of-a-union-to-take)"""
    DICT_WITH_TYPE = 2
    """Instance is a dict with one entry of key type name str and value dict; this serializes to JSON as AVRO expects 
    (except for logical types and bytes) - See:  http://avro.apache.org/docs/current/spec.html#json_encoding"""
    POSSIBLY_AMBIGUOUS_DICT = 3
    """Instance dict provided without type name - which in the case of records with identical fields there is no way
    to determine which is which (example: union of classes A and B where each has identical fields)"""


class AlarmState(Enum):
    NormalDisabled = 1
    """Effectively Normal, out-of-service"""
    NormalFiltered = 2
    """Effectively Normal, suppressed by design"""
    NormalMasked = 3
    """Effectively Normal, hidden by parent alarm"""
    NormalOnDelayed = 4
    """Effectively Normal, temporarily suppressed upon activation"""
    NormalOneShotShelved = 5
    """Effectively Normal, temporarily suppressed until next deactivation or expiration"""
    NormalContinuousShelved = 6
    """Effectively Normal, temporarily suppressed until expiration"""
    ActiveOffDelayed = 7
    """Effectively Active, temporarily incited upon deactivation"""
    ActiveLatched = 8
    """Effectively Active, temporarily incited upon activation"""
    Active = 9
    """Effectively Active, Actually Active, timely operator action required"""
    Normal = 10
    """Effectively Normal, Actually Normal, no action required"""


class OverriddenAlarmType(Enum):
    Disabled = 1
    """A broken alarm can be flagged as out-of-service"""
    Filtered = 2
    """An alarm can be "suppressed by design" - generally a group of alarms are filtered out when not needed for the
    current machine program. The Filter Processor helps operators filter multiple alarms with simple grouping
    commands (like by area)."""
    Masked = 3
    """An alarm can be suppressed by a parent alarm to minimize confusion during an alarm flood and build an 
    alarm hierarchy"""
    OnDelayed = 4
    """An alarm with an on-delay is temporarily suppressed upon activation to minimize fleeting/chattering"""
    OffDelayed = 5
    """An alarm with an off-delay is temporarily incited upon de-activation to minimize fleeting/chattering"""
    Shelved = 6
    """An alarm can be temporarily suppressed via manual operator command"""
    Latched = 7
    """A fleeting alarm (one that toggles between active and not active too quickly) can be configured to require 
    acknowledgement by operators - the alarm is latched once active and won't clear to Normal (or Active) until 
    acknowledged"""


class ShelvedReason(Enum):
    Stale_Alarm = 1
    """Nuisance alarm which remains active for an extended period of time"""
    Chattering_Fleeting_Alarm = 2
    """Nuisance alarm which toggles between active and normal states quickly (fleeting) and may do this often
    (chattering)"""
    Other = 3
    """Some other reason"""


class EPICSSEVR(Enum):
    NO_ALARM = 1
    MINOR = 2
    MAJOR = 3
    INVALID = 4


class EPICSSTAT(Enum):
    NO_ALARM = 1
    READ = 2
    WRITE = 3
    HIHI = 4
    HIGH = 5
    LOLO = 6
    LOW = 7
    STATE = 8
    COS = 9
    COMM = 10
    TIMEOUT = 11
    HW_LIMIT = 12
    CALC = 13
    SCAN = 14
    LINK = 15
    SOFT = 16
    BAD_SUB = 17
    UDF = 18
    DISABLE = 19
    SIMM = 20
    READ_ACCESS = 21
    WRITE_ACCESS = 22


@dataclass
class AlarmLocation:
    """
        Alarm Location record
    """
    parent: Optional[str]
    """Parent Location or null if top-level"""


@dataclass
class SimpleAlarming:
    """
        Simple alarming record (no fields)
    """


@dataclass
class NoteAlarming:
    """
        An alarming record with a note
    """
    note: str
    """A note containing extra information generated at the time of activation"""


@dataclass
class EPICSAlarming:
    """
        An EPICS alarming record
    """
    sevr: EPICSSEVR
    """The severity"""
    stat: EPICSSTAT
    """The status"""


@dataclass
class SimpleProducer:
    """
        Simple alarm producer (no fields)
    """


@dataclass
class EPICSProducer:
    """
        EPICS alarm producer - An alarm producer that produces alarms from EPICS
    """
    pv: str
    """The EPICS Process Variable name"""


@dataclass
class CALCProducer:
    """
        CALC expression alarm producer - An alarm producer that evaluates a CALC expression to produce alarms
    """
    expression: str
    """The CALC (calculate) expression"""


@dataclass
class DisabledOverride:
    """
        Disabled override - Suppresses an alarm that is out-of-service (usually for maintenance)
    """
    comments: Optional[str]
    """Explanation of why the alarm is out-of-service"""


@dataclass
class FilteredOverride:
    """
        Filtered override - Suppresses an alarm via filter rule
    """
    filtername: str
    """Filter rule causing the alarm to be filtered"""


@dataclass
class LatchedOverride:
    """
        Latched override - Incites an alarm until an operator acknowledgement
    """


@dataclass
class MaskedOverride:
    """
        Masked override - Suppresses an alarm when a parent alarm is active
        (establishes a hierarchy and minimizes alarm flooding)
    """


@dataclass
class OnDelayedOverride:
    """
        On-Delay override - Suppresses an alarm for a short duration upon activation
    """
    expiration: int
    """Expiration timestamp (Unix timestamp of milliseconds since Epoch of Jan 1. 1970 UTC)"""


@dataclass
class OffDelayedOverride:
    """
        Off-Delay override - Incites an alarm for a short duration upon deactivation
    """
    expiration: int
    """Expiration timestamp (Unix timestamp of milliseconds since Epoch of Jan 1. 1970 UTC)"""


@dataclass
class ShelvedOverride:
    """
        Shelved override - a temporary override (expires)
    """
    expiration: int
    """Expiration timestamp (Unix timestamp of milliseconds since Epoch of Jan 1. 1970 UTC)"""
    comments: Optional[str]
    """Additional operator comments explaining why the alarm was shelved"""
    reason: ShelvedReason
    """The general motivation for shelving the alarm"""
    oneshot: bool
    """Indicates whether the override expires immediately upon next alarm deactivation 
    (unless timestamp expiration occurs first)"""


@dataclass
class AlarmClass:
    """
        alarm-classes-value subject
    """
    category: str
    """The Alarm Category"""
    priority: AlarmPriority
    """The Alarm Priority"""
    rationale: str
    """The Rationale"""
    corrective_action: str
    """The Corrective Action"""
    point_of_contact_username: str
    """The Point of Contact Username"""
    latching: bool
    """Indicates whether the alarm latches"""
    filterable: bool
    """Indicates whether the alarm can be filtered"""
    on_delay_seconds: Optional[int]
    """(optional) The on-delay in seconds - non-positive is treated as None"""
    off_delay_seconds: Optional[int]
    """(optional) The off-delay in seconds - non-positive is treated as None"""
    screen_path: str
    """The control screen path which provides additional alarm information"""


@dataclass
class AlarmInstance:
    """
        alarm-instances-value subject

        Note: Any attributes inherited from AlarmClass can be set to None which indicate the class value
        should be used.
    """
    alarm_class: str
    """The Alarm Class"""
    producer: Union[SimpleProducer, EPICSProducer, CALCProducer]
    """The Alarm Producer"""
    location: List[str]
    """The Alarm Location"""
    masked_by: Optional[str]
    """(optional) The parent alarm which masks this one"""
    screen_path: str
    """The control screen path which provides additional alarm information"""


@dataclass
class AlarmActivationUnion:
    """
        alarm-activations-value subject
    """
    msg: Union[SimpleAlarming, NoteAlarming, EPICSAlarming]
    """The message payload is a union of possible alarming types"""


@dataclass(frozen=True)
class AlarmOverrideKey:
    """
        alarm-overrides-key subject
    """
    name: str
    """The alarm name"""
    type: OverriddenAlarmType
    """The override type"""


@dataclass
class AlarmOverrideUnion:
    """
        alarm-overrides-value subject
    """
    msg: Union[DisabledOverride, FilteredOverride, LatchedOverride, MaskedOverride, OnDelayedOverride,
               OffDelayedOverride, ShelvedOverride]
    """The message payload is a union of possible override types"""


@dataclass
class AlarmOverrideSet:
    """
        Set of all overrides for an alarm
    """
    disabled: DisabledOverride
    """Disabled Override"""
    filtered: FilteredOverride
    """Filtered Override"""
    latched: LatchedOverride
    """Latched Override"""
    masked: MaskedOverride
    """Masked Override"""
    ondelayed: OnDelayedOverride
    """On Delayed Override"""
    offdelayed: OffDelayedOverride
    """Off Delayed Override"""
    shelved: ShelvedOverride
    """Shelved Override"""


@dataclass
class ProcessorTransitions:
    """
        Set of transition states as alarm data is joined and processed
    """
    transitionToActive: bool
    """true when record is first one to indicate Active after being Normal"""
    transitionToNormal: bool
    """true when record is first one to indicate Normal after being Active"""
    latching: bool
    """true when record is in-process of latching, LatchOverride forthcoming"""
    unshelving: bool
    """true when record is in-process of being unshelved after one-shot, ShelvedOverride clearing"""
    masking: bool
    """true when record is in-process of being masked, MaskedOverride forthcoming"""
    unmasking: bool
    """true when record is in-process of being unmasked, MaskedOverride clearing"""
    ondelaying: bool
    """true when record is in-process of being on-delayed, OnDelayedOverride forthcoming"""
    offdelaying: bool
    """true when record is in-process of being off-delayed, OffDelayedOverride forthcoming"""


@dataclass
class EffectiveRegistration:
    """
        effective-registrations-value subject
    """
    alarm_class: AlarmClass
    """The Alarm Class"""

    actual: AlarmInstance
    """The actual alarm registration instance"""

    calculated: AlarmInstance
    """The calculated alarm registration instance considering class defaults"""


@dataclass
class EffectiveActivation:
    """
        effective-activations-value subject
    """
    actual: AlarmActivationUnion
    """The Alarm Activation"""

    overrides: AlarmOverrideSet
    """The Alarm Overrides"""

    state: AlarmState
    """The calculated AlarmState considering activation and overrides"""


@dataclass
class EffectiveAlarm:
    """
        effective-alarms-value subject
    """
    registration: EffectiveRegistration
    """The effective AlarmRegistration considering class defaults"""

    activation: EffectiveActivation
    """The effective AlarmActivation considering overrides"""


@dataclass
class IntermediateMonolog:
    """
        intermediate-\\*-value subject
    """
    registration: EffectiveRegistration
    """The effective AlarmRegistration considering class defaults"""

    activation: EffectiveActivation
    """The effective AlarmActivation considering overrides"""

    transitions: ProcessorTransitions
    """Transition states used by the alarm processor"""
